#!/usr/bin/perl -w
# nwsip - creates an MDS compliant LDIF out of a NWS subsystem 
# Copyright (C) 2001 by Augusto Ciuffoletti

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
#MA  02111-1307  USA
#
#The author can be contacted by e.mail: augusto@di.unipi.it
######################################
#
#   P A C K A G E    I N T E R V A L
#
######################################
# Measures the delay between the call of the "begin" and "end" method.
# Measures are collected in a array of strings, that is flushed to
# the indicated logfile using the "flush" method
package interval;
use strict;
require 'sys/syscall.ph';

my $x=1234;

sub new {
    shift;
    my ( $logfile ) = @_;
    if ( defined $logfile ) {
	my $state = {};
	$$state{"flag"}=0;
	$$state{"origin"}=0;
	$$state{"title"}="";
	$$state{"report"}=[];
	$$state{"logfile"} = $logfile;
	return bless $state;
    }
    return undef;
}

sub begin {
    my $self = shift;
    my $timeval = pack("LL", ());;

    syscall(&SYS_gettimeofday, $timeval, 0) != -1 
        || die "gettimeofday: $!";
    my @time = unpack("LL",$timeval);
    $$self{"flag"} = 0;
    $$self{"origin"} = $time[0] + ( $time[1] /1000000 );
    $$self{"title"} = $_[0];
    return bless $self;
}

sub end {
    my $self = shift;
    my $timeval = pack("LL", ());

    if ($$self{"flag"} == 0) {
	syscall(&SYS_gettimeofday, $timeval, 0) != -1 
	    || die "gettimeofday: $!";
	my @time = unpack("LL",$timeval);
	$$self{"flag"} = 1;
	my $result = sprintf("%s %f",
			     $$self{"title"},
			     $time[0] + ( $time[1] /1000000 ) - 
			     $$self{"origin"}
			     );
	push @{$$self{"report"}}, $result;
    }
    return;
}

sub flush {
    my $self = shift;
    open LOG, ">> $$self{logfile}" ||
	die "Can't write to logfile\n";
    for ( @{$$self{"report"}} ) {
	print LOG "$_\n";
    }
    $$self{"report"}=[];
}

##################################
#
#     P A C K A G E    S T A N Z A
#
##################################
# Processing of configuration file
package stanza;
use strict;

sub new {
    shift;
    my ( $filename, $name ) = @_;
    my $mapping = {};
    open LINE, $filename
	|| die "Can't open 'foo' for input: $!";
    my $flag=0;
    while ( <LINE> ) {
      SWITCH: {
	  if ( /^#/ || /^\s*$/) { 
	       last SWITCH; 
	   }
	  if ( /^:$name/ ) {
	      $flag=1; 
	      last SWITCH; 
	  } 
	  if ( ! ( /^:$name/ ) && /^:/  ) { 
	      $flag=0; 
	      last SWITCH; 
	  }
	  if ( $flag ) { 
	      my ($in, $out)=split (/\s+/g,$_);
	      $in =~ s/([A-Z])/\l$1/g;
	      $$mapping{$in}=$out;
#test	      print "$in => $$mapping{$in}\n";;
	      last SWITCH;
	  }
      }
    }
    close LINE;
    return bless $mapping;
}

#############
# Given an input and a mapping, represented as a hash, returns the
# string associated to the input in the hash. If the string is "nil"
# returns undefined, and dies if the hash is undefined for that input.
sub map {
    my ( $self, $input ) = @_;;
    my %mapping=%$self;

    $input =~ s/([A-Z])/\l$1/g ;
    exists $mapping{$input} ||
	die "Unknown input name $input: abort\n";
    if ( ! ( $mapping{$input} =~ /nil/ ) ) {
	    return $mapping{$input};
	}
    return;
}

package nwsip;
use strict;

# The directory of nws binaries (configurable global)
my $NWSBINDIR="/usr/local/nws/bin";
# Leave undefined to avoid NWS query delays recording
my $LOGFILE="nwstiming.log";

$|=1;
# The nameserver (global) 
my $nameserver=$ARGV[0]; 
# sensor => ou mapping (global hash) 
my $ou="stanza"->new("nwsip.conf","sensors");
# nws resource to MDS attribute mapping (global hash) 
my $class="stanza"->new("nwsip.conf","classes");
# nws option to MDS attribute mapping (global hash) 
my $option="stanza"->new("nwsip.conf","options");
# nws filed to MDS attribute mapping (global hash) 
my $field="stanza"->new("nwsip.conf","fields");
# A timer 
my $timer = "interval"->new($LOGFILE);
# The commandline
my $cmd="";

# Internal representation of an entry is with a hash

#############
# Reads a nws entry from a nws_search command from a handle
# passed as a parameter and returns an hash (attribute => value)
# Assumes that entries are separated by a single blank line
sub get_entry {
    my $file=shift;
    my %entry=();

    while ( my $attribute = <$file> ){
	chomp $attribute;
	if ( $attribute eq "" ) {
	    last;
	}
	my $name=substr($attribute,0,(index($attribute,':')));
	$name =~ s/\s+$//;
	$entry{$name}=substr($attribute,(index($attribute,':'))+1);
    }
    return %entry;
}

#############
# Given the name of a resource, returns a hash containing the
# attributes coresponding to a measure.

sub format_measures {
    my ( $resource, $host ) = @_;
    my $cmd="$NWSBINDIR/nws_extract -n1 -N$nameserver $resource $host";
    ( defined $timer ) && $timer->begin($cmd);
    open QUERY,"$cmd |"
	|| die "can't fork: $!";
    my @fields = split /\s+/, <QUERY>;
    my @vals = split /\s+/, <QUERY>;
    my $i=0;    
    my %measures=();
    foreach (@fields) {
	( defined $timer ) && $timer->end;
# The "Time" is converted into a readable format 
	if ( $_ eq "Time" ) {
	    open TIME,"$NWSBINDIR/whattime $vals[$i++]|";
	    my ($dontcare,$time) = split /==/,<TIME>;
	    chomp $time;
	    $time =~ s/^\s+//;
	    $measures{$_}=$time;
	    next;
	}
# Other columns are left unchanged
	$measures{$_}=$vals[$i++]; 
    }
    close QUERY;
    return %measures;
}

#############
# Takes a hash describing an NWS series, and a parameter indicating 
# either the local host, or a sensor and a peer, and print as an 
# LDIF entry (except "dn" attribute) the data about the series and
# the last measures.

sub process_NWS_series {
    my ( $refseries, $parameter, $root ) = @_;
    my %series=%$refseries;

# Compute the MDS resource name (to use also as prefix)	
    defined ( my $resource_prefix = 
	         $class->map($series{resource}) ) ||
        return;
    print "dn: $root\n";
    printf "objectclass: NetworkMonitor$resource_prefix\n";
    printf "label: $series{label}\n";
# Output specific attributes
    foreach ( split /,/,$series{option} ) {
	if ( defined ( my $objectclass = $option->map($_) ) ) {
	    $objectclass="$resource_prefix$objectclass";
	    print "objectclass:\tNetworkMonitor$objectclass\n";
	    print "$objectclass:\t$series{$_}\n";
	}
    }

# Output measures for local resources
    my %m=format_measures($series{"resource"}, $parameter);
    foreach my $name ( sort { $b cmp $a } keys %m ) {
	if ( defined ( my $objectclass = $field->map($name) ) ) {
	    printf "objectclass:\tNetworkMonitor$objectclass\n",$resource_prefix;
	    printf "$objectclass:\t$m{$name}\n",$resource_prefix;
	}
    }
    print "\n";
}

#############
# Scan all series registered for a given sensor (1st argument). The 
# most recent  measure of each monitored resource is queried, and 
# a LDIF registration is generated  

sub scan_sensor {
    my ( $ref, $root ) = @_;
    my %sensor=%$ref;
    my %peers=();

# Output the LDIF for this sensor
    print "dn: $root\n";
    print "objectclass: ou\n";
    printf "ou: %s\n",$ou->map($sensor{name});
    print "nwssensor: $sensor{name}\n";
    print "ipaddress: $sensor{ipAddress}\n";
    print "\n";

# First queries the nameserver for all series on this sensor.
    my $cmd="$NWSBINDIR/nws_search -N$nameserver \"&((objectclass=nwsSeries)(host=$sensor{name}))\" ";
    ( defined $timer ) && $timer->begin($cmd);
    open READ_SERIES, "$cmd |"
	|| die "can't fork: $!";
# For each series...
    while ( my %series = get_entry \*READ_SERIES ) {
	( defined $timer ) && $timer->end;
# If the "target" option is present, then treat each (sensor,member)
# pair. A root entry is created that indicates the type of resource,
# and each leaf contains a measure for a peer.
	if ( exists $series{target} ) {
# Push the peer into an array to process later
	    if ( ! exists $peers{$series{target}} ) {
		$peers{$series{target}} = "true";
	    }
	} else {
# Otherwise treat as a local resource
	    process_NWS_series 
		\%series,
		$sensor{name}, 
		"name=$series{name}, $root";
	}
    }
    close READ_SERIES;

# Delete this sensor from peers (just in case...)
    delete $peers{$sensor{name}};

# Scan for resources related to a link
    foreach my $peer ( keys %peers ) {
# Output an entry for the peer
	my $rou=$ou->map($peer);
	print "dn: rou=$rou, $root\n";
	print "objectclass: \tNetworkMonitorOU\n";
	print "nwssensor: \t$peer\n";
	print "rou: \t$rou\n";
	print "\n";
# Reads all series of this sensor related to that peer
	$cmd="$NWSBINDIR/nws_search -N$nameserver \"&((objectclass=nwsSeries)(target=*$peer*))\"";
	( defined $timer ) && $timer->begin($cmd);
	open READ_SERIES, "$cmd |"
	    || die "can't fork: $!";
	while ( my %series = get_entry \*READ_SERIES ) {
	    (defined $timer ) && $timer->end;
	    process_NWS_series 
		\%series, 
		"$sensor{name} $peer", 
		"name=$series{name}, rou=$rou, $root";
	}
	close READ_SERIES;
    }
}

###################
# MAIN PROGRAM
###################
( defined $ARGV[1] ) || die "Usage: $0 <nameserver> <suffix>\n";

# Root nws entry
my $root="cn=nws, $ARGV[1]";

print "dn: $root\n";
print "objectclass: nws\n";
print "nameserver: $nameserver\n";
print "\n";

# Scans all sensor registrations. Each registration is converted into
# an LDIF entry, all series registered for this sensor are scanned

$cmd="$NWSBINDIR/nws_search -N$nameserver \"&((objectclass=nwsHost)(hostType=sensor))\" ";
( defined $timer ) && $timer->begin($cmd);
open READ_SENSORS, "$cmd |"
   || die "can't fork: $!";
while ( my %sensor = get_entry \*READ_SENSORS ) {
    ( defined $timer ) && $timer->end;
    my $rou = $ou->map($sensor{name});
    scan_sensor \%sensor, "ou=$rou, $root";
}
close READ_SENSORS;

( defined $timer ) && $timer->flush();














